"""Tests for DIRECTV."""
import os


def load_fixture(filename):
    """Load a fixture."""
    path = os.path.join(os.path.dirname(__file__), "fixtures", filename)
    with open(path) as fptr:
        return fptr.read()



"""Tests for DirecTV Helpers."""
import directv.utils as utils


def test_combine_channel_number() -> None:
    """Test the merging of channel numbers."""
    assert utils.combine_channel_number(231, 65535) == "231"
    assert utils.combine_channel_number(231, 1) == "231-1"


def test_parse_channel_number() -> None:
    """Test the parsing of channel numbers."""
    assert utils.parse_channel_number("231") == ("231", "65535")



"""Tests for DIRECTV."""
import asyncio

import pytest
from aiohttp import ClientSession
from directv import DIRECTV
from directv.exceptions import (
    DIRECTVAccessRestricted,
    DIRECTVConnectionError,
    DIRECTVError,
)

from . import load_fixture

HOST = "1.2.3.4"
PORT = 8080

MATCH_HOST = f"{HOST}:{PORT}"
NON_STANDARD_PORT = 3333


@pytest.mark.asyncio
async def test_json_request(aresponses):
    """Test DIRECTV response is handled correctly."""
    aresponses.add(
        MATCH_HOST,
        "/info/getVersion",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text='{"status": {"code": 200, "commandResult": 0}}',
        ),
    )

    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        response = await dtv._request("/info/getVersion")
        assert response["status"]["code"] == 200
        assert response["status"]["commandResult"] == 0


@pytest.mark.asyncio
async def test_authenticated_request(aresponses):
    """Test authenticated JSON response is handled correctly."""
    aresponses.add(
        MATCH_HOST,
        "/",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text='{"status": "ok"}',
        ),
    )
    async with ClientSession() as session:
        dtv = DIRECTV(HOST, username="you", password="socool", session=session,)
        response = await dtv._request("/")
        assert response["status"] == "ok"


@pytest.mark.asyncio
async def test_text_request(aresponses):
    """Test non JSON response is handled correctly."""
    aresponses.add(
        MATCH_HOST, "/", "GET", aresponses.Response(status=200, text="OK"),
    )
    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        response = await dtv._request("/")
        assert response == "OK"


@pytest.mark.asyncio
async def test_internal_session(aresponses):
    """Test DIRECTV response is handled correctly."""
    aresponses.add(
        MATCH_HOST,
        "/info/getVersion",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text='{"status": {"code": 200, "commandResult": 0}}',
        ),
    )

    async with DIRECTV(HOST) as dtv:
        response = await dtv._request("/info/getVersion")
        assert response["status"]["code"] == 200
        assert response["status"]["commandResult"] == 0


@pytest.mark.asyncio
async def test_request_port(aresponses):
    """Test the DIRECTV server running on non-standard port."""
    aresponses.add(
        f"{HOST}:{NON_STANDARD_PORT}",
        "/info/getVersion",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text='{"status": {"code": 200, "commandResult": 0}}',
        ),
    )

    async with ClientSession() as session:
        dtv = DIRECTV(host=HOST, port=NON_STANDARD_PORT, session=session,)
        response = await dtv._request("/info/getVersion")
        assert response["status"]["code"] == 200
        assert response["status"]["commandResult"] == 0


@pytest.mark.asyncio
async def test_timeout(aresponses):
    """Test request timeout from the DIRECTV server."""
    # Faking a timeout by sleeping
    async def response_handler(_):
        await asyncio.sleep(2)
        return aresponses.Response(body="Timeout!")

    aresponses.add(
        MATCH_HOST, "/info/getVersion", "GET", response_handler,
    )

    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session, request_timeout=1)
        with pytest.raises(DIRECTVConnectionError):
            assert await dtv._request("/info/getVersion")


@pytest.mark.asyncio
async def test_client_error():
    """Test http client error."""
    async with ClientSession() as session:
        dtv = DIRECTV("#", session=session)
        with pytest.raises(DIRECTVConnectionError):
            assert await dtv._request("/info/getVersion")


@pytest.mark.asyncio
async def test_http_error403(aresponses):
    """Test HTTP 403 response handling."""
    aresponses.add(
        MATCH_HOST,
        "/tv/getTuned",
        "GET",
        aresponses.Response(text="Forbidden", status=403),
    )

    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        with pytest.raises(DIRECTVAccessRestricted):
            assert await dtv._request("/tv/getTuned")


@pytest.mark.asyncio
async def test_http_error404(aresponses):
    """Test HTTP 404 response handling."""
    aresponses.add(
        MATCH_HOST,
        "/info/getVersion",
        "GET",
        aresponses.Response(text="Not Found!", status=404),
    )

    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        with pytest.raises(DIRECTVError):
            assert await dtv._request("/info/getVersion")


@pytest.mark.asyncio
async def test_http_error500(aresponses):
    """Test HTTP 500 response handling."""
    aresponses.add(
        MATCH_HOST,
        "/info/getVersion",
        "GET",
        aresponses.Response(text="Internal Server Error", status=500),
    )

    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        with pytest.raises(DIRECTVError):
            assert await dtv._request("/info/getVersion")


@pytest.mark.asyncio
async def test_http_error500_json(aresponses):
    """Test HTTP 500 json response handling."""
    aresponses.add(
        MATCH_HOST,
        "/info/getVersion",
        "GET",
        aresponses.Response(
            status=500,
            headers={"Content-Type": "application/json"},
            body=load_fixture("info-get-version-error.json"),
        ),
    )

    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        with pytest.raises(DIRECTVError):
            response = await dtv._request("/info/getVersion")
            assert response
            assert response["status"]
            assert response["status"]["code"] == 500
            assert response["status"]["commandResult"] == 1



"""Tests for DIRECTV."""
from typing import List

import pytest
from aiohttp import ClientSession
from directv import DIRECTV, DIRECTVError
from directv.models import Info, Program, State

from . import load_fixture

HOST = "1.2.3.4"
PORT = 8080

MATCH_HOST = f"{HOST}:{PORT}"


@pytest.mark.asyncio
async def test_update(aresponses):
    """Test update is handled correctly."""
    aresponses.add(
        MATCH_HOST,
        "/info/getVersion",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("info-get-version.json"),
        ),
    )

    aresponses.add(
        MATCH_HOST,
        "/info/getLocations",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("info-get-locations.json"),
        ),
    )

    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        response = await dtv.update()

        assert response
        assert isinstance(response.info, Info)
        assert isinstance(response.locations, List)

        response = await dtv.update()

        assert response
        assert response.info


@pytest.mark.asyncio
async def test_remote(aresponses):
    """Test remote is handled correctly."""
    aresponses.add(
        MATCH_HOST,
        "/remote/processKey",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("remote-process-key.json"),
        ),
    )

    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        await dtv.remote("info")


@pytest.mark.asyncio
async def test_remote_invalid_key():
    """Test remote with invalid key is handled correctly."""
    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        with pytest.raises(DIRECTVError):
            await dtv.remote("super")


@pytest.mark.asyncio
async def test_state(aresponses):
    """Test active state is handled correctly."""
    aresponses.add(
        MATCH_HOST,
        "/info/mode",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("info-mode.json"),
        ),
    )

    aresponses.add(
        MATCH_HOST,
        "/tv/getTuned",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("tv-get-tuned.json"),
        ),
    )

    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        response = await dtv.state()

        assert response
        assert isinstance(response, State)
        assert response.available
        assert not response.standby

        assert isinstance(response.program, Program)


@pytest.mark.asyncio
async def test_state_error_mode(aresponses):
    """Test state with generic mode error is handled correctly."""
    aresponses.add(
        MATCH_HOST,
        "/info/mode",
        "GET",
        aresponses.Response(
            status=500,
            headers={"Content-Type": "application/json"},
            text=load_fixture("info-mode-error.json"),
        ),
    )

    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        response = await dtv.state()

        assert response
        assert isinstance(response, State)
        assert not response.available
        assert response.standby
        assert response.authorized

        assert response.program is None


@pytest.mark.asyncio
async def test_state_error_tuned(aresponses):
    """Test state with generic tuned error is handled correctly."""
    aresponses.add(
        MATCH_HOST,
        "/info/mode",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("info-mode.json"),
        ),
    )

    aresponses.add(
        MATCH_HOST,
        "/tv/getTuned",
        "GET",
        aresponses.Response(
            status=500,
            headers={"Content-Type": "application/json"},
            text=load_fixture("tv-get-tuned-error.json"),
        ),
    )

    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        response = await dtv.state()

        assert response
        assert isinstance(response, State)
        assert not response.available
        assert not response.standby
        assert response.authorized

        assert response.program is None


@pytest.mark.asyncio
async def test_state_restricted_mode(aresponses):
    """Test standby state is handled correctly."""
    aresponses.add(
        MATCH_HOST,
        "/info/mode",
        "GET",
        aresponses.Response(
            status=403,
            headers={"Content-Type": "application/json"},
            text=load_fixture("info-mode-restricted.json"),
        ),
    )

    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        response = await dtv.state()

        assert response
        assert isinstance(response, State)
        assert not response.available
        assert response.standby
        assert not response.authorized

        assert response.program is None


@pytest.mark.asyncio
async def test_state_restricted_tuned(aresponses):
    """Test standby state is handled correctly."""
    aresponses.add(
        MATCH_HOST,
        "/info/mode",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("info-mode.json"),
        ),
    )

    aresponses.add(
        MATCH_HOST,
        "/tv/getTuned",
        "GET",
        aresponses.Response(
            status=403,
            headers={"Content-Type": "application/json"},
            text=load_fixture("tv-get-tuned-restricted.json"),
        ),
    )

    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        response = await dtv.state()

        assert response
        assert isinstance(response, State)
        assert response.available
        assert not response.standby
        assert not response.authorized

        assert response.program is None


@pytest.mark.asyncio
async def test_state_standby(aresponses):
    """Test restricted state is handled correctly."""
    aresponses.add(
        MATCH_HOST,
        "/info/mode",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("info-mode-standby.json"),
        ),
    )

    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        response = await dtv.state()

        assert response
        assert isinstance(response, State)
        assert response.available
        assert response.standby

        assert response.program is None


@pytest.mark.asyncio
async def test_status(aresponses):
    """Test active state is handled correctly."""
    aresponses.add(
        MATCH_HOST,
        "/info/mode",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("info-mode.json"),
        ),
    )

    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        response = await dtv.status()

        assert response == "active"


@pytest.mark.asyncio
async def test_status_access_restricted(aresponses):
    """Test unauthorized state is handled correctly."""
    aresponses.add(
        MATCH_HOST,
        "/info/mode",
        "GET",
        aresponses.Response(
            status=403,
            headers={"Content-Type": "application/json"},
            text=load_fixture("info-mode-restricted.json"),
        ),
    )

    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        response = await dtv.status()

        assert response == "unauthorized"


@pytest.mark.asyncio
async def test_status_standby(aresponses):
    """Test standby status is handled correctly."""
    aresponses.add(
        MATCH_HOST,
        "/info/mode",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("info-mode-standby.json"),
        ),
    )

    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        response = await dtv.status()

        assert response == "standby"


@pytest.mark.asyncio
async def test_status_unavailable(aresponses):
    """Test unavailable status is handled correctly."""
    aresponses.add(
        MATCH_HOST,
        "/info/mode",
        "GET",
        aresponses.Response(
            status=500,
            headers={"Content-Type": "application/json"},
            text=load_fixture("info-mode-error.json"),
        ),
    )

    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        response = await dtv.status()

        assert response == "unavailable"


@pytest.mark.asyncio
async def test_tune(aresponses):
    """Test tune is handled correctly."""
    aresponses.add(
        MATCH_HOST,
        "/tv/tune",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("tv-tune.json"),
        ),
    )

    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        await dtv.tune("231")


@pytest.mark.asyncio
async def test_tuned(aresponses):
    """Test tuned is handled correctly."""
    aresponses.add(
        MATCH_HOST,
        "/tv/getTuned",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("tv-get-tuned.json"),
        ),
    )

    async with ClientSession() as session:
        dtv = DIRECTV(HOST, session=session)
        response = await dtv.tuned()

        assert response
        assert isinstance(response, Program)



"""Exceptions for DirecTV."""


class DIRECTVError(Exception):
    """Generic DirecTV exception."""

    pass


class DIRECTVConnectionError(DIRECTVError):
    """DirecTV connection exception."""

    pass


class DIRECTVAccessRestricted(DIRECTVError):
    """DirecTV access restricted."""

    pass



"""Helpers for DirecTV."""
from typing import Tuple


def parse_channel_number(channel: str) -> Tuple[str, str]:
    """Convert a channel number into its major and minor."""
    try:
        major, minor = channel.split("-")
    except ValueError:
        major = channel
        minor = "65535"

    return major, minor


def combine_channel_number(major: int, minor: int) -> str:
    """Create a combined channel number from its major and minor."""
    if minor == 65535:
        return str(major)

    return "%d-%d" % (major, minor)



"""Models for DirecTV."""

from dataclasses import dataclass
from datetime import datetime, timezone
from typing import List, Optional

from .exceptions import DIRECTVError
from .utils import combine_channel_number


@dataclass(frozen=True)
class Info:
    """Object holding information from DirecTV."""

    brand: str
    receiver_id: str
    version: str

    @staticmethod
    def from_dict(data: dict):
        """Return Info object from DirecTV API response."""
        receiver_id = data.get("receiverId", "")

        return Info(
            brand="DirecTV",
            receiver_id="".join(receiver_id.split()),
            version=data.get("stbSoftwareVersion", "Unknown"),
        )


@dataclass(frozen=True)
class Location:
    """Object holding all information of receiver client location."""

    client: bool
    name: str
    address: str

    @staticmethod
    def from_dict(data: dict):
        """Return Info object from DirecTV API response."""
        address = data.get("clientAddr", "")

        return Location(
            client=address != "0",
            name=data.get("locationName", "Receiver"),
            address=address,
        )


@dataclass(frozen=True)
class Program:
    """Object holding all information of playing program."""

    channel: str
    channel_name: str
    ondemand: bool
    recorded: bool
    recording: bool
    viewed: bool
    program_id: int
    program_type: str
    duration: int
    title: str
    episode_title: str
    music_title: str
    music_album: str
    music_artist: str
    partial: bool
    payperview: bool
    position: int
    purchased: bool
    rating: str
    start_time: datetime
    unique_id: int

    @staticmethod
    def from_dict(data: dict):
        """Return Info object from DirecTV API response."""
        major = data.get("major", 0)
        minor = data.get("minor", 65535)
        episode_title = data.get("episodeTitle", None)
        music = data.get("music", {})
        music_title = music.get("title", None)
        program_type = "movie"
        if episode_title is not None:
            program_type = "tvshow"
        elif music_title is not None:
            program_type = "music"
        start_time = data.get("startTime", None)
        if start_time:
            start_time = datetime.fromtimestamp(start_time, timezone.utc)
        unique_id = data.get("uniqueId", None)

        return Program(
            channel=combine_channel_number(major, minor),
            channel_name=data.get("callsign", None),
            program_id=data.get("programId", None),
            program_type=program_type,
            duration=data.get("duration", 0),
            title=data.get("title", None),
            episode_title=episode_title,
            music_title=music_title,
            music_album=music.get("cd", None),
            music_artist=music.get("by", None),
            ondemand=data.get("isVod", False),
            partial=data.get("isPartial", False),
            payperview=data.get("isPpv", False),
            position=data.get("offset", 0),
            purchased=data.get("isPurchased", False),
            rating=data.get("rating", None),
            recorded=(unique_id is not None),
            recording=data.get("isRecording", False),
            start_time=start_time,
            unique_id=unique_id,
            viewed=data.get("isViewed", False),
        )


@dataclass(frozen=True)
class State:
    """Object holding all information of a single receiver client state."""

    authorized: bool
    available: bool
    standby: bool
    program: Optional[Program]
    at: datetime = datetime.utcnow()


class Device:
    """Object holding all information of receiver."""

    info: Info
    locations: List[Location] = []

    def __init__(self, data: dict):
        """Initialize an empty DirecTV device class."""
        # Check if all elements are in the passed dict, else raise an Error
        if any(k not in data for k in ["locations", "info"]):
            raise DIRECTVError(
                "DirecTV data is incomplete, cannot construct device object"
            )
        self.update_from_dict(data)

    def update_from_dict(self, data: dict) -> "Device":
        """Return Device object from DirecTV API response."""
        if "info" in data and data["info"]:
            self.info = Info.from_dict(data["info"])

        if "locations" in data and data["locations"]:
            locations = [Location.from_dict(location) for location in data["locations"]]
            self.locations = locations

        return self



"""Asynchronous Python client for DirecTV."""
import asyncio
import json
from socket import gaierror as SocketGIAEroor
from typing import Any, Mapping, Optional

import aiohttp
import async_timeout
from yarl import URL

from .__version__ import __version__
from .const import VALID_REMOTE_KEYS
from .exceptions import DIRECTVAccessRestricted, DIRECTVConnectionError, DIRECTVError
from .models import Device, Program, State
from .utils import parse_channel_number


class DIRECTV:
    """Main class for handling connections with DirecTV servers."""

    _device: Optional[Device] = None

    def __init__(
        self,
        host: str,
        base_path: str = "/",
        password: str = None,
        port: int = 8080,
        request_timeout: int = 8,
        session: aiohttp.client.ClientSession = None,
        username: str = None,
        user_agent: str = None,
    ) -> None:
        """Initialize connection with receiver."""
        self._session = session
        self._close_session = False

        self.base_path = base_path
        self.host = host
        self.password = password
        self.port = port
        self.request_timeout = request_timeout
        self.username = username
        self.user_agent = user_agent

        if user_agent is None:
            self.user_agent = f"PythonDirecTV/{__version__}"

    async def _request(
        self,
        uri: str = "",
        method: str = "GET",
        data: Optional[Any] = None,
        params: Optional[Mapping[str, str]] = None,
    ) -> Any:
        """Handle a request to a receiver."""
        scheme = "http"

        url = URL.build(
            scheme=scheme, host=self.host, port=self.port, path=self.base_path
        ).join(URL(uri))

        auth = None
        if self.username and self.password:
            auth = aiohttp.BasicAuth(self.username, self.password)

        headers = {
            "User-Agent": self.user_agent,
            "Accept": "application/json, text/plain, */*",
        }

        if self._session is None:
            self._session = aiohttp.ClientSession()
            self._close_session = True

        try:
            with async_timeout.timeout(self.request_timeout):
                response = await self._session.request(
                    method, url, auth=auth, data=data, params=params, headers=headers,
                )
        except asyncio.TimeoutError as exception:
            raise DIRECTVConnectionError(
                "Timeout occurred while connecting to receiver"
            ) from exception
        except (aiohttp.ClientError, SocketGIAEroor) as exception:
            raise DIRECTVConnectionError(
                "Error occurred while communicating with receiver"
            ) from exception

        if response.status == 403:
            raise DIRECTVAccessRestricted(
                "Access restricted. Please ensure external device access is allowed",
                {},
            )

        content_type = response.headers.get("Content-Type")

        if (response.status // 100) in [4, 5]:
            content = await response.read()
            response.close()

            if content_type == "application/json":
                raise DIRECTVError(
                    f"HTTP {response.status}", json.loads(content.decode("utf8"))
                )

            raise DIRECTVError(
                f"HTTP {response.status}",
                {
                    "content-type": content_type,
                    "message": content.decode("utf8"),
                    "status-code": response.status,
                },
            )

        if "application/json" in content_type:
            data = await response.json()
            return data

        return await response.text()

    @property
    def device(self) -> Optional[Device]:
        """Return the cached Device object."""
        return self._device

    async def update(self, full_update: bool = False) -> Device:
        """Get all information about the device in a single call."""
        if self._device is None or full_update:
            info = await self._request("info/getVersion")
            if info is None:
                raise DIRECTVError("DirecTV device returned an empty API response")

            locations = await self._request("info/getLocations")
            if locations is None or "locations" not in locations:
                raise DIRECTVError("DirecTV device returned an empty API response")

            self._device = Device({"info": info, "locations": locations["locations"]})
            return self._device

        self._device.update_from_dict({})
        return self._device

    async def remote(self, key: str, client: str = "0") -> None:
        """Emulate pressing a key on the remote.

        Supported keys: power, poweron, poweroff, format,
        pause, rew, replay, stop, advance, ffwd, record,
        play, guide, active, list, exit, back, menu, info,
        up, down, left, right, select, red, green, yellow,
        blue, chanup, chandown, prev, 0, 1, 2, 3, 4, 5,
        6, 7, 8, 9, dash, enter
        """
        if not key.lower() in VALID_REMOTE_KEYS:
            raise DIRECTVError(f"Remote key is invalid: {key}")

        keypress = {
            "key": key,
            "hold": "keyPress",
            "clientAddr": client,
        }

        await self._request("remote/processKey", params=keypress)

    async def state(self, client: str = "0") -> State:
        """Get state of receiver client."""
        authorized = True
        program = None

        try:
            mode = await self._request("info/mode", params={"clientAddr": client})
            available = True
            standby = mode["mode"] == 1
        except DIRECTVAccessRestricted:
            authorized = False
            available = False
            standby = True
        except DIRECTVError:
            available = False
            standby = True

        if not standby:
            try:
                program = await self.tuned(client)
            except DIRECTVAccessRestricted:
                authorized = False
                program = None
            except DIRECTVError:
                available = False
                program = None

        return State(
            authorized=authorized,
            available=available,
            standby=standby,
            program=program,
        )

    async def status(self, client: str = "0") -> str:
        """Get basic status of receiver client."""
        try:
            mode = await self._request("info/mode", params={"clientAddr": client})
            return "standby" if mode["mode"] == 1 else "active"
        except DIRECTVAccessRestricted:
            return "unauthorized"
        except DIRECTVError:
            return "unavailable"

    async def tune(self, channel: str, client: str = "0") -> None:
        """Change the channel on the receiver."""
        major, minor = parse_channel_number(channel)

        tune = {
            "major": major,
            "minor": minor,
            "clientAddr": client,
        }

        await self._request("tv/tune", params=tune)

    async def tuned(self, client: str = "0") -> Program:
        """Get currently tuned program."""
        tuned = await self._request("tv/getTuned", params={"clientAddr": client})
        return Program.from_dict(tuned)

    async def close(self) -> None:
        """Close open client session."""
        if self._session and self._close_session:
            await self._session.close()

    async def __aenter__(self) -> "DIRECTV":
        """Async enter."""
        return self

    async def __aexit__(self, *exc_info) -> None:
        """Async exit."""
        await self.close()



"""Constants for DirecTV."""

VALID_REMOTE_KEYS = [
    "power",
    "poweron",
    "poweroff",
    "format",
    "pause",
    "rew",
    "replay",
    "stop",
    "advance",
    "ffwd",
    "record",
    "play",
    "guide",
    "active",
    "list",
    "exit",
    "back",
    "menu",
    "info",
    "up",
    "down",
    "left",
    "right",
    "select",
    "red",
    "green",
    "yellow",
    "blue",
    "chanup",
    "chandown",
    "prev",
    "0",
    "1",
    "2",
    "3",
    "4",
    "5",
    "6",
    "7",
    "8",
    "9",
    "dash",
    "enter",
]



"""Tests for DirecTV Models."""
from datetime import datetime, timezone

import directv.models as models
import pytest
from directv import DIRECTVError

INFO = {
    "accessCardId": "0021-1495-6572",
    "receiverId": "0288 7745 5858",
    "stbSoftwareVersion": "0x4ed7",
    "systemTime": 1281625203,
    "version": "1.2",
}

LOCATIONS = [
    {"clientAddr": "0", "locationName": "Host"},
    {"clientAddr": "2CA17D1CD30X", "locationName": "Client"},
]

DEVICE = {"info": INFO, "locations": LOCATIONS}

PROGRAM = {
    "callsign": "FOODHD",
    "date": "20070324",
    "duration": 1791,
    "episodeTitle": "Spaghetti and Clam Sauce",
    "expiration": "0",
    "expiryTime": 0,
    "isOffAir": False,
    "isPartial": False,
    "isPclocked": 1,
    "isPpv": False,
    "isRecording": False,
    "isViewed": True,
    "isVod": False,
    "keepUntilFull": True,
    "major": 231,
    "minor": 65535,
    "offset": 263,
    "programId": "4405732",
    "rating": "No Rating",
    "recType": 3,
    "startTime": 1278342008,
    "stationId": 3900976,
    "title": "Tyler's Ultimate",
    "uniqueId": "6728716739474078694",
}

PROGRAM_MOVIE = {
    "callsign": "HALLHD",
    "date": "2013",
    "duration": 7200,
    "isOffAir": False,
    "isPclocked": 3,
    "isPpv": False,
    "isRecording": False,
    "isVod": False,
    "major": 312,
    "minor": 65535,
    "offset": 4437,
    "programId": "17016356",
    "rating": "TV-G",
    "startTime": 1584795600,
    "stationId": 6580971,
    "title": "Snow Bride",
}

PROGRAM_MUSIC = {
    "callsign": "MCSJ",
    "duration": 86400,
    "isOffAir": False,
    "isPclocked": 3,
    "isPpv": False,
    "isRecording": False,
    "isVod": False,
    "major": 851,
    "minor": 65535,
    "music": {
        "by": "Gerald Albright",
        "cd": "Slam Dunk (2014)",
        "title": "Sparkle In Your Eyes",
    },
    "offset": 15050,
    "programId": "76917562",
    "rating": "TV-PG",
    "startTime": 1584784800,
    "stationId": 2872196,
    "title": "Smooth Jazz",
}


def test_device() -> None:
    """Test the Device model."""
    device = models.Device(DEVICE)

    assert device

    assert device.info
    assert isinstance(device.info, models.Info)

    assert device.locations
    assert len(device.locations) == 2
    assert isinstance(device.locations[0], models.Location)


def test_device_no_data() -> None:
    """Test the Device model."""
    with pytest.raises(DIRECTVError):
        models.Device({})


def test_info() -> None:
    """Test the Info model."""
    info = models.Info.from_dict(INFO)

    assert info
    assert info.brand == "DirecTV"
    assert info.version == "0x4ed7"
    assert info.receiver_id == "028877455858"


def test_location() -> None:
    """Test the Location model."""
    location = models.Location.from_dict(LOCATIONS[0])

    assert location
    assert not location.client
    assert location.name == "Host"
    assert location.address == "0"

    location = models.Location.from_dict(LOCATIONS[1])

    assert location
    assert location.client
    assert location.name == "Client"
    assert location.address == "2CA17D1CD30X"


def test_program() -> None:
    """Test the Program model."""
    program = models.Program.from_dict(PROGRAM)

    assert program
    assert program.recorded
    assert program.viewed
    assert not program.ondemand
    assert not program.partial
    assert not program.payperview
    assert not program.purchased
    assert not program.recording
    assert program.channel == "231"
    assert program.channel_name == "FOODHD"
    assert program.program_id == "4405732"
    assert program.program_type == "tvshow"
    assert program.title == "Tyler's Ultimate"
    assert program.episode_title == "Spaghetti and Clam Sauce"
    assert program.rating == "No Rating"
    assert program.start_time == datetime(2010, 7, 5, 15, 0, 8, tzinfo=timezone.utc)
    assert program.duration == 1791
    assert program.position == 263
    assert program.unique_id == "6728716739474078694"


def test_program_movie() -> None:
    """Test the Program model with movie."""
    program = models.Program.from_dict(PROGRAM_MOVIE)

    assert program
    assert not program.recorded
    assert not program.viewed
    assert not program.ondemand
    assert not program.partial
    assert not program.payperview
    assert not program.purchased
    assert not program.recording
    assert program.channel == "312"
    assert program.channel_name == "HALLHD"
    assert program.program_id == "17016356"
    assert program.program_type == "movie"
    assert program.title == "Snow Bride"
    assert program.episode_title is None
    assert program.rating == "TV-G"
    assert program.start_time == datetime(2020, 3, 21, 13, 0, tzinfo=timezone.utc)
    assert program.duration == 7200
    assert program.position == 4437
    assert program.unique_id is None


def test_program_music() -> None:
    """Test the Program model with music channel."""
    program = models.Program.from_dict(PROGRAM_MUSIC)

    assert program
    assert not program.recorded
    assert not program.viewed
    assert not program.ondemand
    assert not program.partial
    assert not program.payperview
    assert not program.purchased
    assert not program.recording
    assert program.channel == "851"
    assert program.channel_name == "MCSJ"
    assert program.program_id == "76917562"
    assert program.program_type == "music"
    assert program.title == "Smooth Jazz"
    assert program.episode_title is None
    assert program.music_title == "Sparkle In Your Eyes"
    assert program.music_album == "Slam Dunk (2014)"
    assert program.music_artist == "Gerald Albright"
    assert program.rating == "TV-PG"
    assert program.start_time == datetime(2020, 3, 21, 10, 0, tzinfo=timezone.utc)
    assert program.duration == 86400
    assert program.position == 15050
    assert program.unique_id is None


def test_state() -> None:
    """Test the State model."""
    program = models.Program.from_dict(PROGRAM)
    state = models.State(
        authorized=True, available=True, standby=False, program=program,
    )

    assert state
    assert isinstance(state.at, datetime)
